/****************************************************************************
** Meta object code from reading C++ file 'TerminalSession.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../contour/src/contour/TerminalSession.h"
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'TerminalSession.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN7contour15TerminalSessionE_t {};
} // unnamed namespace

template <> constexpr inline auto contour::TerminalSession::qt_create_metaobjectdata<qt_meta_tag_ZN7contour15TerminalSessionE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "contour::TerminalSession",
        "sessionClosed",
        "",
        "TerminalSession&",
        "profileNameChanged",
        "newValue",
        "lineCountChanged",
        "columnsCountChanged",
        "historyLineCountChanged",
        "scrollOffsetChanged",
        "titleChanged",
        "value",
        "pathToBackgroundChanged",
        "opacityBackgroundChanged",
        "isImageBackgroundChanged",
        "isBlurBackgroundChanged",
        "backgroundColorChanged",
        "isScrollbarRightChanged",
        "isScrollbarVisibleChanged",
        "opacityChanged",
        "onBell",
        "volume",
        "onAlert",
        "requestPermissionForFontChange",
        "requestPermissionForPasteLargeFile",
        "requestPermissionForBufferCapture",
        "requestPermissionForShowHostWritableStatusLine",
        "showNotification",
        "title",
        "content",
        "fontSizeChanged",
        "createNewTab",
        "closeTab",
        "switchToPreviousTab",
        "switchToTabLeft",
        "switchToTabRight",
        "switchToTab",
        "position",
        "onConfigReload",
        "onHighlightUpdate",
        "configureDisplay",
        "applyPendingFontChange",
        "allow",
        "remember",
        "applyPendingPaste",
        "executePendingBufferCapture",
        "executeShowHostWritableStatusLine",
        "resizeTerminalToDisplaySize",
        "id",
        "pageLineCount",
        "pageColumnsCount",
        "showResizeIndicator",
        "historyLineCount",
        "scrollOffset",
        "opacity",
        "pathToBackground",
        "opacityBackground",
        "isImageBackground",
        "isBlurBackground",
        "backgroundColor",
        "QColor",
        "isScrollbarRight",
        "isScrollbarVisible",
        "fontSize",
        "upTime",
        "bellSource"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'sessionClosed'
        QtMocHelpers::SignalData<void(TerminalSession &)>(1, 2, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 3, 2 },
        }}),
        // Signal 'profileNameChanged'
        QtMocHelpers::SignalData<void(QString)>(4, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 5 },
        }}),
        // Signal 'lineCountChanged'
        QtMocHelpers::SignalData<void(int)>(6, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 5 },
        }}),
        // Signal 'columnsCountChanged'
        QtMocHelpers::SignalData<void(int)>(7, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 5 },
        }}),
        // Signal 'historyLineCountChanged'
        QtMocHelpers::SignalData<void(int)>(8, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 5 },
        }}),
        // Signal 'scrollOffsetChanged'
        QtMocHelpers::SignalData<void(int)>(9, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 5 },
        }}),
        // Signal 'titleChanged'
        QtMocHelpers::SignalData<void(QString const &)>(10, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 11 },
        }}),
        // Signal 'pathToBackgroundChanged'
        QtMocHelpers::SignalData<void()>(12, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'opacityBackgroundChanged'
        QtMocHelpers::SignalData<void()>(13, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'isImageBackgroundChanged'
        QtMocHelpers::SignalData<void()>(14, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'isBlurBackgroundChanged'
        QtMocHelpers::SignalData<void()>(15, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'backgroundColorChanged'
        QtMocHelpers::SignalData<void()>(16, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'isScrollbarRightChanged'
        QtMocHelpers::SignalData<void()>(17, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'isScrollbarVisibleChanged'
        QtMocHelpers::SignalData<void()>(18, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'opacityChanged'
        QtMocHelpers::SignalData<void()>(19, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'onBell'
        QtMocHelpers::SignalData<void(float)>(20, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Float, 21 },
        }}),
        // Signal 'onAlert'
        QtMocHelpers::SignalData<void()>(22, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestPermissionForFontChange'
        QtMocHelpers::SignalData<void()>(23, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestPermissionForPasteLargeFile'
        QtMocHelpers::SignalData<void()>(24, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestPermissionForBufferCapture'
        QtMocHelpers::SignalData<void()>(25, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'requestPermissionForShowHostWritableStatusLine'
        QtMocHelpers::SignalData<void()>(26, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'showNotification'
        QtMocHelpers::SignalData<void(QString const &, QString const &)>(27, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 28 }, { QMetaType::QString, 29 },
        }}),
        // Signal 'fontSizeChanged'
        QtMocHelpers::SignalData<void()>(30, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'createNewTab'
        QtMocHelpers::SignalData<void()>(31, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'closeTab'
        QtMocHelpers::SignalData<void()>(32, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'switchToPreviousTab'
        QtMocHelpers::SignalData<void()>(33, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'switchToTabLeft'
        QtMocHelpers::SignalData<void()>(34, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'switchToTabRight'
        QtMocHelpers::SignalData<void()>(35, 2, QMC::AccessPublic, QMetaType::Void),
        // Signal 'switchToTab'
        QtMocHelpers::SignalData<void(int)>(36, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 37 },
        }}),
        // Slot 'onConfigReload'
        QtMocHelpers::SlotData<void()>(38, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onHighlightUpdate'
        QtMocHelpers::SlotData<void()>(39, 2, QMC::AccessPublic, QMetaType::Void),
        // Slot 'configureDisplay'
        QtMocHelpers::SlotData<void()>(40, 2, QMC::AccessPublic, QMetaType::Void),
        // Method 'applyPendingFontChange'
        QtMocHelpers::MethodData<void(bool, bool)>(41, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 42 }, { QMetaType::Bool, 43 },
        }}),
        // Method 'applyPendingPaste'
        QtMocHelpers::MethodData<void(bool, bool)>(44, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 42 }, { QMetaType::Bool, 43 },
        }}),
        // Method 'executePendingBufferCapture'
        QtMocHelpers::MethodData<void(bool, bool)>(45, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 42 }, { QMetaType::Bool, 43 },
        }}),
        // Method 'executeShowHostWritableStatusLine'
        QtMocHelpers::MethodData<void(bool, bool)>(46, 2, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Bool, 42 }, { QMetaType::Bool, 43 },
        }}),
        // Method 'resizeTerminalToDisplaySize'
        QtMocHelpers::MethodData<void()>(47, 2, QMC::AccessPublic, QMetaType::Void),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'id'
        QtMocHelpers::PropertyData<int>(48, QMetaType::Int, QMC::DefaultPropertyFlags),
        // property 'pageLineCount'
        QtMocHelpers::PropertyData<int>(49, QMetaType::Int, QMC::DefaultPropertyFlags, 2),
        // property 'pageColumnsCount'
        QtMocHelpers::PropertyData<int>(50, QMetaType::Int, QMC::DefaultPropertyFlags, 3),
        // property 'showResizeIndicator'
        QtMocHelpers::PropertyData<bool>(51, QMetaType::Bool, QMC::DefaultPropertyFlags),
        // property 'historyLineCount'
        QtMocHelpers::PropertyData<int>(52, QMetaType::Int, QMC::DefaultPropertyFlags, 4),
        // property 'scrollOffset'
        QtMocHelpers::PropertyData<int>(53, QMetaType::Int, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 5),
        // property 'title'
        QtMocHelpers::PropertyData<QString>(28, QMetaType::QString, QMC::DefaultPropertyFlags | QMC::Writable | QMC::StdCppSet, 6),
        // property 'opacity'
        QtMocHelpers::PropertyData<float>(54, QMetaType::Float, QMC::DefaultPropertyFlags, 14),
        // property 'pathToBackground'
        QtMocHelpers::PropertyData<QString>(55, QMetaType::QString, QMC::DefaultPropertyFlags, 7),
        // property 'opacityBackground'
        QtMocHelpers::PropertyData<float>(56, QMetaType::Float, QMC::DefaultPropertyFlags, 8),
        // property 'isImageBackground'
        QtMocHelpers::PropertyData<bool>(57, QMetaType::Bool, QMC::DefaultPropertyFlags, 9),
        // property 'isBlurBackground'
        QtMocHelpers::PropertyData<bool>(58, QMetaType::Bool, QMC::DefaultPropertyFlags, 10),
        // property 'backgroundColor'
        QtMocHelpers::PropertyData<QColor>(59, 0x80000000 | 60, QMC::DefaultPropertyFlags | QMC::EnumOrFlag, 11),
        // property 'isScrollbarRight'
        QtMocHelpers::PropertyData<bool>(61, QMetaType::Bool, QMC::DefaultPropertyFlags, 12),
        // property 'isScrollbarVisible'
        QtMocHelpers::PropertyData<bool>(62, QMetaType::Bool, QMC::DefaultPropertyFlags, 13),
        // property 'fontSize'
        QtMocHelpers::PropertyData<int>(63, QMetaType::Int, QMC::DefaultPropertyFlags),
        // property 'upTime'
        QtMocHelpers::PropertyData<int>(64, QMetaType::Int, QMC::DefaultPropertyFlags),
        // property 'bellSource'
        QtMocHelpers::PropertyData<QString>(65, QMetaType::QString, QMC::DefaultPropertyFlags, 15),
    };
    QtMocHelpers::UintData qt_enums {
    };
    return QtMocHelpers::metaObjectData<TerminalSession, qt_meta_tag_ZN7contour15TerminalSessionE_t>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums);
}
Q_CONSTINIT const QMetaObject contour::TerminalSession::staticMetaObject = { {
    QMetaObject::SuperData::link<QAbstractItemModel::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN7contour15TerminalSessionE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN7contour15TerminalSessionE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN7contour15TerminalSessionE_t>.metaTypes,
    nullptr
} };

void contour::TerminalSession::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<TerminalSession *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->sessionClosed((*reinterpret_cast<std::add_pointer_t<TerminalSession&>>(_a[1]))); break;
        case 1: _t->profileNameChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 2: _t->lineCountChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 3: _t->columnsCountChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 4: _t->historyLineCountChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 5: _t->scrollOffsetChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 6: _t->titleChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 7: _t->pathToBackgroundChanged(); break;
        case 8: _t->opacityBackgroundChanged(); break;
        case 9: _t->isImageBackgroundChanged(); break;
        case 10: _t->isBlurBackgroundChanged(); break;
        case 11: _t->backgroundColorChanged(); break;
        case 12: _t->isScrollbarRightChanged(); break;
        case 13: _t->isScrollbarVisibleChanged(); break;
        case 14: _t->opacityChanged(); break;
        case 15: _t->onBell((*reinterpret_cast<std::add_pointer_t<float>>(_a[1]))); break;
        case 16: _t->onAlert(); break;
        case 17: _t->requestPermissionForFontChange(); break;
        case 18: _t->requestPermissionForPasteLargeFile(); break;
        case 19: _t->requestPermissionForBufferCapture(); break;
        case 20: _t->requestPermissionForShowHostWritableStatusLine(); break;
        case 21: _t->showNotification((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 22: _t->fontSizeChanged(); break;
        case 23: _t->createNewTab(); break;
        case 24: _t->closeTab(); break;
        case 25: _t->switchToPreviousTab(); break;
        case 26: _t->switchToTabLeft(); break;
        case 27: _t->switchToTabRight(); break;
        case 28: _t->switchToTab((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 29: _t->onConfigReload(); break;
        case 30: _t->onHighlightUpdate(); break;
        case 31: _t->configureDisplay(); break;
        case 32: _t->applyPendingFontChange((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 33: _t->applyPendingPaste((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 34: _t->executePendingBufferCapture((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 35: _t->executeShowHostWritableStatusLine((*reinterpret_cast<std::add_pointer_t<bool>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<bool>>(_a[2]))); break;
        case 36: _t->resizeTerminalToDisplaySize(); break;
        default: ;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)(TerminalSession & )>(_a, &TerminalSession::sessionClosed, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)(QString )>(_a, &TerminalSession::profileNameChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)(int )>(_a, &TerminalSession::lineCountChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)(int )>(_a, &TerminalSession::columnsCountChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)(int )>(_a, &TerminalSession::historyLineCountChanged, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)(int )>(_a, &TerminalSession::scrollOffsetChanged, 5))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)(QString const & )>(_a, &TerminalSession::titleChanged, 6))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::pathToBackgroundChanged, 7))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::opacityBackgroundChanged, 8))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::isImageBackgroundChanged, 9))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::isBlurBackgroundChanged, 10))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::backgroundColorChanged, 11))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::isScrollbarRightChanged, 12))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::isScrollbarVisibleChanged, 13))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::opacityChanged, 14))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)(float )>(_a, &TerminalSession::onBell, 15))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::onAlert, 16))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::requestPermissionForFontChange, 17))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::requestPermissionForPasteLargeFile, 18))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::requestPermissionForBufferCapture, 19))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::requestPermissionForShowHostWritableStatusLine, 20))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)(QString const & , QString const & )>(_a, &TerminalSession::showNotification, 21))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::fontSizeChanged, 22))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::createNewTab, 23))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::closeTab, 24))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::switchToPreviousTab, 25))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::switchToTabLeft, 26))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)()>(_a, &TerminalSession::switchToTabRight, 27))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalSession::*)(int )>(_a, &TerminalSession::switchToTab, 28))
            return;
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<int*>(_v) = _t->id(); break;
        case 1: *reinterpret_cast<int*>(_v) = _t->pageLineCount(); break;
        case 2: *reinterpret_cast<int*>(_v) = _t->pageColumnsCount(); break;
        case 3: *reinterpret_cast<bool*>(_v) = _t->showResizeIndicator(); break;
        case 4: *reinterpret_cast<int*>(_v) = _t->historyLineCount(); break;
        case 5: *reinterpret_cast<int*>(_v) = _t->scrollOffset(); break;
        case 6: *reinterpret_cast<QString*>(_v) = _t->title(); break;
        case 7: *reinterpret_cast<float*>(_v) = _t->getOpacity(); break;
        case 8: *reinterpret_cast<QString*>(_v) = _t->pathToBackground(); break;
        case 9: *reinterpret_cast<float*>(_v) = _t->getOpacityBackground(); break;
        case 10: *reinterpret_cast<bool*>(_v) = _t->getIsImageBackground(); break;
        case 11: *reinterpret_cast<bool*>(_v) = _t->getIsBlurBackground(); break;
        case 12: *reinterpret_cast<QColor*>(_v) = _t->getBackgroundColor(); break;
        case 13: *reinterpret_cast<bool*>(_v) = _t->getIsScrollbarRight(); break;
        case 14: *reinterpret_cast<bool*>(_v) = _t->getIsScrollbarVisible(); break;
        case 15: *reinterpret_cast<int*>(_v) = _t->getFontSize(); break;
        case 16: *reinterpret_cast<int*>(_v) = _t->getUptime(); break;
        case 17: *reinterpret_cast<QString*>(_v) = _t->getBellSource(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 5: _t->setScrollOffset(*reinterpret_cast<int*>(_v)); break;
        case 6: _t->setTitle(*reinterpret_cast<QString*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *contour::TerminalSession::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *contour::TerminalSession::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN7contour15TerminalSessionE_t>.strings))
        return static_cast<void*>(this);
    if (!strcmp(_clname, "vtbackend::Terminal::Events"))
        return static_cast< vtbackend::Terminal::Events*>(this);
    return QAbstractItemModel::qt_metacast(_clname);
}

int contour::TerminalSession::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QAbstractItemModel::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 37)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 37;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 37)
            *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType();
        _id -= 37;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 18;
    }
    return _id;
}

// SIGNAL 0
void contour::TerminalSession::sessionClosed(TerminalSession & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 0, nullptr, _t1);
}

// SIGNAL 1
void contour::TerminalSession::profileNameChanged(QString _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void contour::TerminalSession::lineCountChanged(int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void contour::TerminalSession::columnsCountChanged(int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void contour::TerminalSession::historyLineCountChanged(int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 4, nullptr, _t1);
}

// SIGNAL 5
void contour::TerminalSession::scrollOffsetChanged(int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1);
}

// SIGNAL 6
void contour::TerminalSession::titleChanged(QString const & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 6, nullptr, _t1);
}

// SIGNAL 7
void contour::TerminalSession::pathToBackgroundChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 7, nullptr);
}

// SIGNAL 8
void contour::TerminalSession::opacityBackgroundChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 8, nullptr);
}

// SIGNAL 9
void contour::TerminalSession::isImageBackgroundChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 9, nullptr);
}

// SIGNAL 10
void contour::TerminalSession::isBlurBackgroundChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 10, nullptr);
}

// SIGNAL 11
void contour::TerminalSession::backgroundColorChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 11, nullptr);
}

// SIGNAL 12
void contour::TerminalSession::isScrollbarRightChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 12, nullptr);
}

// SIGNAL 13
void contour::TerminalSession::isScrollbarVisibleChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 13, nullptr);
}

// SIGNAL 14
void contour::TerminalSession::opacityChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 14, nullptr);
}

// SIGNAL 15
void contour::TerminalSession::onBell(float _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 15, nullptr, _t1);
}

// SIGNAL 16
void contour::TerminalSession::onAlert()
{
    QMetaObject::activate(this, &staticMetaObject, 16, nullptr);
}

// SIGNAL 17
void contour::TerminalSession::requestPermissionForFontChange()
{
    QMetaObject::activate(this, &staticMetaObject, 17, nullptr);
}

// SIGNAL 18
void contour::TerminalSession::requestPermissionForPasteLargeFile()
{
    QMetaObject::activate(this, &staticMetaObject, 18, nullptr);
}

// SIGNAL 19
void contour::TerminalSession::requestPermissionForBufferCapture()
{
    QMetaObject::activate(this, &staticMetaObject, 19, nullptr);
}

// SIGNAL 20
void contour::TerminalSession::requestPermissionForShowHostWritableStatusLine()
{
    QMetaObject::activate(this, &staticMetaObject, 20, nullptr);
}

// SIGNAL 21
void contour::TerminalSession::showNotification(QString const & _t1, QString const & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 21, nullptr, _t1, _t2);
}

// SIGNAL 22
void contour::TerminalSession::fontSizeChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 22, nullptr);
}

// SIGNAL 23
void contour::TerminalSession::createNewTab()
{
    QMetaObject::activate(this, &staticMetaObject, 23, nullptr);
}

// SIGNAL 24
void contour::TerminalSession::closeTab()
{
    QMetaObject::activate(this, &staticMetaObject, 24, nullptr);
}

// SIGNAL 25
void contour::TerminalSession::switchToPreviousTab()
{
    QMetaObject::activate(this, &staticMetaObject, 25, nullptr);
}

// SIGNAL 26
void contour::TerminalSession::switchToTabLeft()
{
    QMetaObject::activate(this, &staticMetaObject, 26, nullptr);
}

// SIGNAL 27
void contour::TerminalSession::switchToTabRight()
{
    QMetaObject::activate(this, &staticMetaObject, 27, nullptr);
}

// SIGNAL 28
void contour::TerminalSession::switchToTab(int _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 28, nullptr, _t1);
}
QT_WARNING_POP
