/****************************************************************************
** Meta object code from reading C++ file 'TerminalDisplay.h'
**
** Created by: The Qt Meta Object Compiler version 69 (Qt 6.11.0)
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include "../../../../../../contour/src/contour/display/TerminalDisplay.h"
#include <QtNetwork/QSslPreSharedKeyAuthenticator>
#include <QtNetwork/QSslError>
#include <QtCore/qmetatype.h>

#include <QtCore/qtmochelpers.h>

#include <memory>


#include <QtCore/qxptype_traits.h>
#if !defined(Q_MOC_OUTPUT_REVISION)
#error "The header file 'TerminalDisplay.h' doesn't include <QObject>."
#elif Q_MOC_OUTPUT_REVISION != 69
#error "This file was generated using the moc from 6.11.0. It"
#error "cannot be used with the include files from this version of Qt."
#error "(The moc has changed too much.)"
#endif

#ifndef Q_CONSTINIT
#define Q_CONSTINIT
#endif

QT_WARNING_PUSH
QT_WARNING_DISABLE_DEPRECATED
QT_WARNING_DISABLE_GCC("-Wuseless-cast")
namespace {
struct qt_meta_tag_ZN7contour7display15TerminalDisplayE_t {};
} // unnamed namespace

template <> constexpr inline auto contour::display::TerminalDisplay::qt_create_metaobjectdata<qt_meta_tag_ZN7contour7display15TerminalDisplayE_t>()
{
    namespace QMC = QtMocConstants;
    QtMocHelpers::StringRefStorage qt_stringData {
        "contour::display::TerminalDisplay",
        "QML.Element",
        "auto",
        "profileNameChanged",
        "",
        "titleChanged",
        "sessionChanged",
        "TerminalSession*",
        "terminalBufferChanged",
        "vtbackend::ScreenType",
        "terminated",
        "showNotification",
        "title",
        "body",
        "onAutoScrollTick",
        "onSceneGrapheInitialized",
        "onBeforeSynchronize",
        "onBeforeRendering",
        "paint",
        "handleWindowChanged",
        "QQuickWindow*",
        "newWindow",
        "sizeChanged",
        "cleanup",
        "onAfterRendering",
        "onScrollBarValueChanged",
        "value",
        "onRefreshRateChanged",
        "applyFontDPI",
        "onScreenChanged",
        "onDpiConfigChanged",
        "doDumpState",
        "logDisplayInfo",
        "pageLineCount",
        "historyLineCount",
        "session",
        "profile"
    };

    QtMocHelpers::UintData qt_methods {
        // Signal 'profileNameChanged'
        QtMocHelpers::SignalData<void()>(3, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'titleChanged'
        QtMocHelpers::SignalData<void(QString const &)>(5, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 4 },
        }}),
        // Signal 'sessionChanged'
        QtMocHelpers::SignalData<void(TerminalSession *)>(6, 4, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 7, 4 },
        }}),
        // Signal 'terminalBufferChanged'
        QtMocHelpers::SignalData<void(vtbackend::ScreenType)>(8, 4, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 9, 4 },
        }}),
        // Signal 'terminated'
        QtMocHelpers::SignalData<void()>(10, 4, QMC::AccessPublic, QMetaType::Void),
        // Signal 'showNotification'
        QtMocHelpers::SignalData<void(QString const &, QString const &)>(11, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::QString, 12 }, { QMetaType::QString, 13 },
        }}),
        // Slot 'onAutoScrollTick'
        QtMocHelpers::SlotData<void()>(14, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onSceneGrapheInitialized'
        QtMocHelpers::SlotData<void()>(15, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onBeforeSynchronize'
        QtMocHelpers::SlotData<void()>(16, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onBeforeRendering'
        QtMocHelpers::SlotData<void()>(17, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'paint'
        QtMocHelpers::SlotData<void()>(18, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'handleWindowChanged'
        QtMocHelpers::SlotData<void(QQuickWindow *)>(19, 4, QMC::AccessPublic, QMetaType::Void, {{
            { 0x80000000 | 20, 21 },
        }}),
        // Slot 'sizeChanged'
        QtMocHelpers::SlotData<void()>(22, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'cleanup'
        QtMocHelpers::SlotData<void()>(23, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onAfterRendering'
        QtMocHelpers::SlotData<void()>(24, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onScrollBarValueChanged'
        QtMocHelpers::SlotData<void(int)>(25, 4, QMC::AccessPublic, QMetaType::Void, {{
            { QMetaType::Int, 26 },
        }}),
        // Slot 'onRefreshRateChanged'
        QtMocHelpers::SlotData<void()>(27, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'applyFontDPI'
        QtMocHelpers::SlotData<void()>(28, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onScreenChanged'
        QtMocHelpers::SlotData<void()>(29, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'onDpiConfigChanged'
        QtMocHelpers::SlotData<void()>(30, 4, QMC::AccessPublic, QMetaType::Void),
        // Slot 'doDumpState'
        QtMocHelpers::SlotData<void()>(31, 4, QMC::AccessPublic, QMetaType::Void),
        // Method 'logDisplayInfo'
        QtMocHelpers::MethodData<void()>(32, 4, QMC::AccessPublic, QMetaType::Void),
        // Method 'pageLineCount'
        QtMocHelpers::MethodData<int() const>(33, 4, QMC::AccessPublic, QMetaType::Int),
        // Method 'historyLineCount'
        QtMocHelpers::MethodData<int() const>(34, 4, QMC::AccessPublic, QMetaType::Int),
    };
    QtMocHelpers::UintData qt_properties {
        // property 'session'
        QtMocHelpers::PropertyData<TerminalSession*>(35, 0x80000000 | 7, QMC::DefaultPropertyFlags | QMC::Writable | QMC::EnumOrFlag | QMC::StdCppSet, 2),
        // property 'profile'
        QtMocHelpers::PropertyData<QString>(36, QMetaType::QString, QMC::DefaultPropertyFlags | QMC::Writable, 0),
        // property 'title'
        QtMocHelpers::PropertyData<QString>(12, QMetaType::QString, QMC::DefaultPropertyFlags, 1),
    };
    QtMocHelpers::UintData qt_enums {
    };
    QtMocHelpers::UintData qt_constructors {};
    QtMocHelpers::ClassInfos qt_classinfo({
            {    1,    2 },
    });
    return QtMocHelpers::metaObjectData<TerminalDisplay, void>(QMC::MetaObjectFlag{}, qt_stringData,
            qt_methods, qt_properties, qt_enums, qt_constructors, qt_classinfo);
}
Q_CONSTINIT const QMetaObject contour::display::TerminalDisplay::staticMetaObject = { {
    QMetaObject::SuperData::link<QQuickItem::staticMetaObject>(),
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN7contour7display15TerminalDisplayE_t>.stringdata,
    qt_staticMetaObjectStaticContent<qt_meta_tag_ZN7contour7display15TerminalDisplayE_t>.data,
    qt_static_metacall,
    nullptr,
    qt_staticMetaObjectRelocatingContent<qt_meta_tag_ZN7contour7display15TerminalDisplayE_t>.metaTypes,
    nullptr
} };

void contour::display::TerminalDisplay::qt_static_metacall(QObject *_o, QMetaObject::Call _c, int _id, void **_a)
{
    auto *_t = static_cast<TerminalDisplay *>(_o);
    if (_c == QMetaObject::InvokeMetaMethod) {
        switch (_id) {
        case 0: _t->profileNameChanged(); break;
        case 1: _t->titleChanged((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1]))); break;
        case 2: _t->sessionChanged((*reinterpret_cast<std::add_pointer_t<TerminalSession*>>(_a[1]))); break;
        case 3: _t->terminalBufferChanged((*reinterpret_cast<std::add_pointer_t<vtbackend::ScreenType>>(_a[1]))); break;
        case 4: _t->terminated(); break;
        case 5: _t->showNotification((*reinterpret_cast<std::add_pointer_t<QString>>(_a[1])),(*reinterpret_cast<std::add_pointer_t<QString>>(_a[2]))); break;
        case 6: _t->onAutoScrollTick(); break;
        case 7: _t->onSceneGrapheInitialized(); break;
        case 8: _t->onBeforeSynchronize(); break;
        case 9: _t->onBeforeRendering(); break;
        case 10: _t->paint(); break;
        case 11: _t->handleWindowChanged((*reinterpret_cast<std::add_pointer_t<QQuickWindow*>>(_a[1]))); break;
        case 12: _t->sizeChanged(); break;
        case 13: _t->cleanup(); break;
        case 14: _t->onAfterRendering(); break;
        case 15: _t->onScrollBarValueChanged((*reinterpret_cast<std::add_pointer_t<int>>(_a[1]))); break;
        case 16: _t->onRefreshRateChanged(); break;
        case 17: _t->applyFontDPI(); break;
        case 18: _t->onScreenChanged(); break;
        case 19: _t->onDpiConfigChanged(); break;
        case 20: _t->doDumpState(); break;
        case 21: _t->logDisplayInfo(); break;
        case 22: { int _r = _t->pageLineCount();
            if (_a[0]) *reinterpret_cast<int*>(_a[0]) = std::move(_r); }  break;
        case 23: { int _r = _t->historyLineCount();
            if (_a[0]) *reinterpret_cast<int*>(_a[0]) = std::move(_r); }  break;
        default: ;
        }
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        switch (_id) {
        default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
        case 2:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< TerminalSession* >(); break;
            }
            break;
        case 11:
            switch (*reinterpret_cast<int*>(_a[1])) {
            default: *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType(); break;
            case 0:
                *reinterpret_cast<QMetaType *>(_a[0]) = QMetaType::fromType< QQuickWindow* >(); break;
            }
            break;
        }
    }
    if (_c == QMetaObject::IndexOfMethod) {
        if (QtMocHelpers::indexOfMethod<void (TerminalDisplay::*)()>(_a, &TerminalDisplay::profileNameChanged, 0))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalDisplay::*)(QString const & )>(_a, &TerminalDisplay::titleChanged, 1))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalDisplay::*)(TerminalSession * )>(_a, &TerminalDisplay::sessionChanged, 2))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalDisplay::*)(vtbackend::ScreenType )>(_a, &TerminalDisplay::terminalBufferChanged, 3))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalDisplay::*)()>(_a, &TerminalDisplay::terminated, 4))
            return;
        if (QtMocHelpers::indexOfMethod<void (TerminalDisplay::*)(QString const & , QString const & )>(_a, &TerminalDisplay::showNotification, 5))
            return;
    }
    if (_c == QMetaObject::RegisterPropertyMetaType) {
        switch (_id) {
        default: *reinterpret_cast<int*>(_a[0]) = -1; break;
        case 0:
            *reinterpret_cast<int*>(_a[0]) = qRegisterMetaType< TerminalSession* >(); break;
        }
    }
    if (_c == QMetaObject::ReadProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: *reinterpret_cast<TerminalSession**>(_v) = _t->getSessionHelper(); break;
        case 1: *reinterpret_cast<QString*>(_v) = _t->profileName(); break;
        case 2: *reinterpret_cast<QString*>(_v) = _t->title(); break;
        default: break;
        }
    }
    if (_c == QMetaObject::WriteProperty) {
        void *_v = _a[0];
        switch (_id) {
        case 0: _t->setSession(*reinterpret_cast<TerminalSession**>(_v)); break;
        case 1: _t->setProfileName(*reinterpret_cast<QString*>(_v)); break;
        default: break;
        }
    }
}

const QMetaObject *contour::display::TerminalDisplay::metaObject() const
{
    return QObject::d_ptr->metaObject ? QObject::d_ptr->dynamicMetaObject() : &staticMetaObject;
}

void *contour::display::TerminalDisplay::qt_metacast(const char *_clname)
{
    if (!_clname) return nullptr;
    if (!strcmp(_clname, qt_staticMetaObjectStaticContent<qt_meta_tag_ZN7contour7display15TerminalDisplayE_t>.strings))
        return static_cast<void*>(this);
    return QQuickItem::qt_metacast(_clname);
}

int contour::display::TerminalDisplay::qt_metacall(QMetaObject::Call _c, int _id, void **_a)
{
    _id = QQuickItem::qt_metacall(_c, _id, _a);
    if (_id < 0)
        return _id;
    if (_c == QMetaObject::InvokeMetaMethod) {
        if (_id < 24)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 24;
    }
    if (_c == QMetaObject::RegisterMethodArgumentMetaType) {
        if (_id < 24)
            qt_static_metacall(this, _c, _id, _a);
        _id -= 24;
    }
    if (_c == QMetaObject::ReadProperty || _c == QMetaObject::WriteProperty
            || _c == QMetaObject::ResetProperty || _c == QMetaObject::BindableProperty
            || _c == QMetaObject::RegisterPropertyMetaType) {
        qt_static_metacall(this, _c, _id, _a);
        _id -= 3;
    }
    return _id;
}

// SIGNAL 0
void contour::display::TerminalDisplay::profileNameChanged()
{
    QMetaObject::activate(this, &staticMetaObject, 0, nullptr);
}

// SIGNAL 1
void contour::display::TerminalDisplay::titleChanged(QString const & _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 1, nullptr, _t1);
}

// SIGNAL 2
void contour::display::TerminalDisplay::sessionChanged(TerminalSession * _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 2, nullptr, _t1);
}

// SIGNAL 3
void contour::display::TerminalDisplay::terminalBufferChanged(vtbackend::ScreenType _t1)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 3, nullptr, _t1);
}

// SIGNAL 4
void contour::display::TerminalDisplay::terminated()
{
    QMetaObject::activate(this, &staticMetaObject, 4, nullptr);
}

// SIGNAL 5
void contour::display::TerminalDisplay::showNotification(QString const & _t1, QString const & _t2)
{
    QMetaObject::activate<void>(this, &staticMetaObject, 5, nullptr, _t1, _t2);
}
QT_WARNING_POP
